#encoding: UTF-8

=begin
RestrictedEntry is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

RestrictedEntry is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with RestrictedEntry; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

Copyright 2014 Detlef Reichl  detlef!reichl()gmx!org
=end


require 'i18n-toolbox'

# Rgw::RestrictionEntry is a specialized Gtk::Entry, that can be restricted so that only valid
# Numerics can be entered. For this the locale under which the program runs is taken into account.
# Also the range of the input can be limited. For numerics it will limit the number range, for
# strings the length of the in put.

module Rgw
    class RestrictedEntry < Gtk::Entry
        alias_method :text_parent, :text
        # Initialize a new Rgw::RestrictedEntry object
        def initialize
            super()
            @restriction = :none
            @lcNumeric = I18nToolbox::LcNumeric.new
            @beepOnError = true
            @isValid = false
            @range = nil
            
            @matchInteger = Regexp.new /[0-9-]/
            @matchFloat = Regexp.new /[0-9#{@lcNumeric.delimiter.chr}-]/
            @matchFloatString = Regexp.new /^\-?[0-9]*#{@lcNumeric.delimiter.chr}{,1}[0-9]*$/
            
            signal_connect(:insert_text) {|wid, str| signal_emit_stop :insert_text unless chr_valid? str}
            signal_connect_after(:insert_text) {str_valid?}
            signal_connect_after(:delete_text) {str_valid?}
        end
        
        
        # Sets the restriction to the entry. Only characters which match the selected type will show
        # up in the entry.
        #
        # @param rest [:integer, :float, :none] The entry will be limited to characters to display
        #        Integers (:integer), Floats (:float) or has no restriction (:none)
        def restriction=(rest)
            raise ArgumentError, "invalid type %s for restriction" % rest.class.to_s unless rest.is_a? Symbol
            raise ArgumentError, "invalid restriction %s" % rest.to_s unless [:integer, :float, :none].include? rest
            @restriction = rest
        end


        # Set the range of valid values for the entry
        #
        # param rg [Range] for a restriction of :integer or :float the range gives the minimum and
        #       maximum values, which are allowed.
        #       for a restriction of :none the range gives the minimum or maximum length of the
        #       inserted string
        def range=(rg)
            raise ArgumentError, "invalid type %s for range" % rg.class.to_s unless rg.is_a? Range
            raise ArgumentError, "range has to be numeric" unless rg.first.is_a? Numeric and rg.last.is_a? Numeric
            raise ArgumentError, "range start has to be positive for none-restrictions" if @restriction == :none and rg.first < 0
            @range = rg
        end 

        
        # Enables to beep if the string does not represent a valid number, i.e. has two decimal
        # points or is empty
        #
        # @param beep [true, false] if beeping should be enabled
        def beep_on_error=(beep)
            @beepOnError = beep == true
        end

        
        # Returns the text of the entry
        #
        # @return [String, nil] Returns the string, if it is valid or nil if it is not.
        def text
            return super() if str_valid?
            nil
        end
        

private
        # Checks if the text in the entry is a valid numeric representation
        def str_valid?
            txt = self.text_parent
            if @restriction == :none
                unless @range.nil?
                   @isValid = @range.cover? txt.length
                else
                    @isValid = true
                end
            else
                @isValid = 0 < txt.length
                @isValid &= @matchFloatString =~ txt
                @isValid &= @range.cover? @lcNumeric.to_f txt  unless @range.nil?
            end
            @isValid ? set_secondary_icon_stock('') : set_secondary_icon_stock('gtk-dialog-warning')
            Gdk.beep if !@isValid and @beepOnError
            @isValid
        end


        # Checks if a single character is valid for the selected restriction
        def chr_valid? str
            return true if @restriction == :none
            
            if @restriction == :integer
                return @matchInteger =~ str
            end
            
            if @restriction == :float
                return @matchFloat =~ str
            end
        end
    end
end
